# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Mix:
   """
   mix class.
   
   Attributes
   ----------
   alpha : float
      Mixing parameter.
   imem : integer
      Mixer memory.
   maxit : integer
      Maximal number of self-consistent iterations.
   method : string
      Mixer algorithm.
   metric : string
      Mixer metric.
   precond : string
      Mixer preconditioner.
   tol : float
      Mixer tolerance.
   type : string
      Mixer type (e.g. density or potential).

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "alpha" : False,
         "precond" : False,
         "metric" : False,
         "imem" : False,
         "method" : False,
         "tol" : False,
         "type" : False,
         "maxit" : False,
      }
      # default values
      defDict = {
         "alpha":0.2,
         "precond":"eye",
         "metric":"eye",
         "imem":20,
         "method":"pul",
         "tol":1E-8,
         "type":"den",
         "maxit":100
      }
      
      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)


# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Mpidist:
   """
   mpi class.
   
   Attributes
   ----------
   grdblk : integer
      Grid blocking factor.
   kptblk : integer
      K-point blocking factor.
   orbblk : integer
      Orbital (matrices) blocking factor.
   bndblk : integer
      Band blocking factor.
   nrgblk : integer
      Energy blocking factor.
   imgblk : integer
      Image (NEB) blocking factor.

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "bndblk": False,
         "bndprc": False,
         "grdblk": False,
         "grdprc": False,
         "imgblk": False,      
         "imgprc": False,      
         "kptblk": False,
         "kptprc": False,
         "nrgblk": False,
         "nrgprc": False,
         "orbblk": False,
         "orbprc": False,
      }    
      # default values
      defDict = {
         "bndblk": 4,
         "bndprc": None,
         "grdblk": 5,
         "grdprc": None,
         "imgblk": 1,
         "imgprc": None,      
         "kptblk": 1,
         "kptprc": None,
         "nrgblk": 1,
         "nrgprc": None,
         "orbblk": 32,
         "orbprc": None,
      }    
      
      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)


# -*- coding: utf-8 -*-
"""
Created on 2020-06-16

@author: Vincent Michaud-Rioux
"""

import numpy as np
import h5py
from matplotlib import pyplot as plt
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import set_units as utils_set_units

class Dos:
   """
   dos class.
   
   Attributes
   ----------
   interval : 1D array
      Interval on which the DOS is calculated (e.g. [-5.5,8.1])
   resolution : float
      Grid resolution for the DOS interval

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "dos"             : False,
         "dosUnits"        : False,
         "efermi"          : False,
         "efermiUnits"     : False,
         "energy"          : False,
         "energyUnits"     : False,
         "interval"        : False,
         "intervalUnits"   : False,
         "lpdos"           : False,
         "orbA"            : False,
         "orbL"            : False,
         "orbM"            : False,         
         "pdos"            : False,
         "pdosUnits"       : False,
         "resolution"      : False,
         "resolutionUnits" : False,
      }
      # default values
      defDict = {
         "dos"             : None,
         "dosUnits"        : "invev",
         "efermi"          : None,
         "efermiUnits"     : "ev",
         "energy"          : None,
         "energyUnits"     : "ev",
         "interval"        : [-10.0, 10.0],
         "intervalUnits"   : "ev",
         "lpdos"           : False,
         "orbA"            : None,
         "orbL"            : None,
         "orbM"            : None,         
         "pdos"            : None,
         "pdosUnits"       : "invev",
         "resolution"      : 0.025,
         "resolutionUnits" : "ev",
      }    

      init_from_dict(self, reqDict, defDict, inpDict)

      for key in ["dos", "energy", "orbA", "orbL", "orbM", "pdos"]:
         attr = getattr(self, key)
         if isinstance(attr, list):
            setattr(self, key, np.array(attr))

      if isinstance(self.pdos, str):
         f = h5py.File(self.pdos, "r")
         self.pdos = f["dos"]["pdos"][0:].T

   def asdict(self):
      return utils_asdict(self)

   def plot_dos(self):
      x = self.energy
      y = self.dos
      fig = plt.figure()
      plt.plot(x, y, "-k")
      plt.xlabel("Energy ("+self.energyUnits+")")
      plt.ylabel("DOS ("+self.dosUnits+")")
      plt.show()
      return fig

   def plot_pdos(self, sumA=None, sumL=None, sumM=None):
      if sum([sumA is None, sumL is None, sumM is None]) > 1:
         NameError("User may not specify two sumX parameters.")
      x = self.energy
      y = self.dos
      fig = plt.figure()
      plt.plot(x, y, "-k")
      if all([sumA is None, sumL is None, sumM is None]):
         plt.show()
      else:
         y = self.pdos.reshape((x.size,-1), order='F')
         if sumA is None: sumA = [sumA]
         if sumL is None: sumL = [sumL]
         if sumM is None: sumM = [sumM]
         plot_pdos_core(x,y,[[self.orbA,sumA,"A"],[self.orbL,sumL,"L"],[self.orbM,sumM,"M"]])
      plt.xlabel("Energy ("+self.energyUnits+")")
      plt.ylabel("DOS ("+self.dosUnits+")")
      plt.show()
      return fig

   def set_units(self, units):
      return utils_set_units(self, units)

def plot_pdos_core(x, pdos, orbXXX):
   for a in orbXXX[0][1]:
      if a is None:
         orbA = np.ones_like(orbXXX[0][0], dtype=bool)
         labA = []
      else:
         orbA = orbXXX[0][0] == a
         labA = [orbXXX[0][2]," = ",str(a),","]
      for l in orbXXX[1][1]:
         if l is None:
            orbL = np.ones_like(orbXXX[1][0], dtype=bool)
            labL = []
         else:
            orbL = orbXXX[1][0] == l
            labL = [orbXXX[1][2]," = ",str(l),","]
         for m in orbXXX[2][1]:
            if m is None:
               orbM = np.ones_like(orbXXX[2][0], dtype=bool)
               labM = []
            else:
               orbM = orbXXX[2][0] == m
               labM = [orbXXX[2][2]," = ",str(m),","]
            orbX = np.logical_and(orbA, orbL)
            orbX = np.logical_and(orbX, orbM)
            d = np.sum(pdos[:,orbX], axis=1)
            label = "".join(labA+labL+labM)[0:-1]
            plt.plot(x, d,"-",label=label)
            plt.legend(loc="upper right")



# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.cell    import bra2avec
from rescupy.utils   import asdict as utils_asdict
from rescupy.utils   import init_from_dict
from rescupy.utils   import set_units as utils_set_units

class Kmesh:
   """
   kmesh class.
   
   Attributes
   ----------
   bvec : 2D array
      Reciprocal unit kmesh row-vectors
   kpts : 2D array
      k-points in fractional coordinates
   kwght : 1D array
      k-points weights

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):

      # required status
      reqDict = {
         "bvec"         : False,
         "bvecUnits"    : False,
         "checkPoints"  : False,
         "gammaCentered": False,
         "grid"         : False,
         "kpts"         : False,
         "kwght"        : False,
         "type"         : False,
      }
      # default values
      defDict = {
         "bvec"         : None,
         "bvecUnits"    : "invang",
         "checkPoints"  : None,
         "gammaCentered": False,
         "grid"         : None,
         "kpts"         : None,
         "kwght"        : None,
         "type"         : "full",
      }  

      init_from_dict(self, reqDict, defDict, inpDict)

      if not self.grid is None and self.kpts is None:
         if self.type == "full": 
            kpts = monkhorst_pack(self.grid)
            if self.gammaCentered:
               shift = np.array(self.grid)
               shift = np.mod(shift + 1, 2) / 2. / shift
               kpts = kpts + shift.reshape((-1,3))
            self.kpts = kpts
            nkpt = np.prod(self.grid)
            self.kwght = np.ones(nkpt) / nkpt

      if not self.kpts is None and len(self.kwght) < len(self.kpts) / 3:
         raise NameError('kwght and kpts have incompatible dimensions.')

      if not self.kwght is None and not np.isclose(1., np.sum(self.kwght)):
         raise NameError('kwght does not sum to 1.')

      if isinstance(self.kpts, list):
         self.kpts = np.array(self.kpts).reshape((-1,3))

   def asdict(self):
      return utils_asdict(self)

   def get_kpts_num(self):
      if isinstance(self.kpts, list):
         nkpt = len(self.kpts) // 3
      elif isinstance(self.kpts, np.ndarray):
         nkpt = self.kpts.shape[0]
      else:
         raise NameError('Invalid type for attribute kpts.')
      return nkpt

   def get_bz_dist(self, kpt0, kpt1):
      dist = np.matmul(kpt1 - kpt0, self.bvec)
      return np.linalg.norm(dist)

   def get_bz_segment(self, kpt0, kpt1, n = None, res = None):
      if n is None and res is None:
         NameError("n or res must be specified.")
      if not n is None:
         kpts = np.zeros((n, 3))
         kpts[:,0] = np.linspace(kpt0[0], kpt1[0], n)
         kpts[:,1] = np.linspace(kpt0[1], kpt1[1], n)
         kpts[:,2] = np.linspace(kpt0[2], kpt1[2], n)
      else: 
         dist = self.get_bz_dist(kpt0, kpt1)
         n = np.ceil(dist / res).astype(int)
         kpts = self.get_bz_segment(kpt0, kpt1, n = n)
      return kpts

   def kpt_2_label(self, cell, kpts):
      if cell.bravais is None:
         cell.bravais = cell.get_bravais_lattice()
      bravais = cell.bravais
      avec = cell.avec
      avecStd = cell.get_standard_lattice()
      pmat = np.matmul(avecStd, np.linalg.inv(avec)).transpose()
      labels = []
      for i in range(0, len(kpts)):
         k = np.matmul(kpts[i],pmat)
         labels.append(kValue2kSymbol(bravais, avecStd, k))
      return labels

   def set_units(self, units):
      return utils_set_units(self, units)

   def set_bvec(self, cell):
      self.bvec, self.bvecUnits = cell.reciprocal()

   def set_bz_lines(self, cell):
      if not self.kpts is None:
         return
      if cell.bravais is None:
         cell.bravais = cell.get_bravais_lattice()
      bravais = cell.bravais
      avec = cell.avec
      avecStd = cell.get_standard_lattice()
      pmat = np.matmul(avec, np.linalg.inv(avecStd)).transpose()
      # define checkpoints
      if self.checkPoints is None:
         checkPoints = defaultSymmetryKPoints(bravais, avecStd)
      else:
         checkPoints = self.checkPoints
      # get line segments
      checkPoints = split_bz_lines(checkPoints)
      # convert to numerical format
      for i in range(0, len(checkPoints)):
         kpt0 = kSymbol2kValue(bravais, avecStd, checkPoints[i][0]) # checkPoint returned as is if not a str
         if isinstance(checkPoints[i][0], str): kpt0 = np.matmul(kpt0,pmat)
         kpt1 = kSymbol2kValue(bravais, avecStd, checkPoints[i][1]) # checkPoint returned as is if not a str
         if isinstance(checkPoints[i][1], str): kpt1 = np.matmul(kpt1,pmat)
         checkPoints[i] = [kpt0, kpt1]
      # get total length
      klen = 0.0
      for i in range(0, len(checkPoints)):
         kpt0 = checkPoints[i][0]
         kpt1 = checkPoints[i][1]
         klen += self.get_bz_dist(kpt0, kpt1)
      # get resolution
      if not self.grid is None:
         res = klen / self.grid
      else:
         res = 0.015
      # interpolate between checkpoints to obtain segments
      kpts = np.zeros((0,3))
      count = 0; checkIndex = np.zeros(2*len(checkPoints)).astype(int)
      for i in range(0, len(checkPoints)):
         kpt0 = checkPoints[i][0]
         kpt1 = checkPoints[i][1]
         ktmp = self.get_bz_segment(kpt0, kpt1, res = res)
         kpts = np.concatenate((kpts, ktmp), axis = 0)
         checkIndex[2*i] = count
         checkIndex[2*i+1] = count + ktmp.shape[0] - 1
         count = checkIndex[2*i+1] + 1
      self.kpts = kpts
      nkpt = kpts.shape[0]
      self.kwght = [1. / nkpt for i in range(nkpt)]
      self.grid = [nkpt]
      self.checkPoints = checkIndex

####################
# module functions #
####################

def monkhorst_pack(grid):
   kx = monkhorst_pack_nodes(grid[0])
   ky = monkhorst_pack_nodes(grid[1])
   kz = monkhorst_pack_nodes(grid[2])
   kx, ky, kz = np.meshgrid(kx, ky, kz, indexing='ij')
   kpts = np.hstack((kx.reshape((-1,1), order='F'), 
                     ky.reshape((-1,1), order='F'), 
                     kz.reshape((-1,1), order='F')))
   return kpts

def monkhorst_pack_nodes(n):
   k = np.linspace(0., 1., num=n+1)[0:n]
   if (n % 2 == 0):
      k += .5 / n
   k -= np.round(k)
   return k

def split_bz_lines(checkPoints):
   if not isinstance(checkPoints[0], list):
      return split_list(checkPoints)
   newlist = []
   for i in range(0, len(checkPoints)):
      newlist = newlist+split_list(checkPoints[i])
   return newlist
         
def split_list(alist):
   newlist = []
   for i in range(0, len(alist) - 1):
      newlist.append(alist[i:i + 2])
   return newlist

def increase_ksampling(k, kref):
   k0 = k
   kw = np.array(kref).astype(float)
   while all([a >= b for a, b in zip(k0, k)]):
      kw *= 1.1
      k = np.round(kw).astype(int).tolist()
   return k

def defaultSymmetryKPoints(crystalStructure, avecStd):
   """
   Return a linear BZ sampling given the crystal structure and lattice vectors.

   Parameters
   ----------
   crystalStructure : string
      Name of the Bravais lattice.

   avecStd : 2D array
      3 x 3 array of row-vectors corresponding to the lattice vectors in the standard 
      ordering (AFLOW).

   Returns
   -------
   symmetryKPoints : list of strings
      Path of high-symmetry k-point labels.

   Author Info:
   Lei Liu, HZWTech
   2015-10
   Vincent Michaud-Rioux
   2020-06-23
   """

   avecStd = avecStd.transpose() # here primitiveCellVectors = [a1 a2 a3]
   epsilon = 1e-3
   if crystalStructure.lower() in ["cubic", "cub", "simplecubic", "sc", "cp"]:
      # 1: ucv = pcv = [[1 0 0]', [0 1 0]', [0 0 1]']
      # symmetryKPoints = [["G","X","M","G","R","X"],["M","R"]]
      symmetryKPoints = [["G","X","M","G","R","X"],["M","R"]]
      # ["G", "X", "M", "G", "R", "X"] ["M", "R"]
   elif crystalStructure.lower() in ["facecenteredcubic", "fcc", "cf"]:
      # 2: pcv = [[0 1./2. 1./2.]', [1./2., 0, 1./2.]', [1./2. 1./2. 0]']
      #    ucv = [[1 0 0]', [0 1 0]', [0 0 1]']
      # symmetryKPoints = ["U","X","G","X","W","K","G","L","U","W","L","K"]
      symmetryKPoints = [["G","X","W","K","G","L","U","W","L","K"],["U","X"]]
      # ["G","X","W","K","G","L","U","W","L","K"] ["U","X"]
   elif crystalStructure.lower() in ["bodycenteredcubic", "bcc", "ci"]:
      # 3: pcv = [[-1./2. 1./2. 1./2.]', [1./2., -1./2., 1./2.]', [1./2. 1./2. -1./2.]']
      #    ucv = [[1 0 0]', [0 1 0]', [0 0 1]']
      # symmetryKPoints = [["G","H","N","G","P","H"] ["P","N"]]
      symmetryKPoints = [["G","H","N","G","P","H"],["P","N"]]
      # ["G","H","N","G","P","H"] ["P","N"]
   elif crystalStructure.lower() in ["tetragonal", "tet", "tp"]:
      # 4: ucv = pcv = [[1 0 0]', [0 1 0]', [0 0 5]']
      # symmetryKPoints = [["G", "X", "M", "G", "Z", "R", "A", "Z"] ["X", "R", "G", "A", "M"]]
      symmetryKPoints = [["G","X","M","G","Z","R","A","Z"],["X","R"],["M","A"]]
      # ["G", "X", "M", "G", "Z", "R", "A", "Z"] ["X", "R"]  ["M", "A"]
   elif crystalStructure.lower() in ["bodycenteredtetragonal", "bct", "ti"]:
      # 5: pcv = [[-1./2. 1./2. 5/2]', [1./2., -1./2., 5/2]', [1./2. 1./2. -5/2]']
      #    ucv = [[1 0 0]', [0 1 0]', [0 0 5]']
      pcv4bcc = np.array([[-1./2.,1./2.,1./2.], [1./2., -1./2., 1./2.], [1./2., 1./2., -1./2.]]).transpose()
      ucv4bcc = np.eye(3)
      pcv = avecStd
      ucv = pcv*(np.linalg.inv(pcv4bcc)*ucv4bcc)
      if np.norm(ucv[:,2]) < np.norm(ucv[:,0]):
         type = "1"
      else:
         type = "2"
      if type is "1":
         # symmetryKPoints = ["X", "P", "G", "X", "M", "G", "Z", "P", "N", "Z1", "M"]
         symmetryKPoints = [["G","X","M","G","Z","P","N","Z1","M"],["X","P"]]
         # ["G", "X", "M", "G", "Z", "P", "N", "Z1", "M"] ["X", "P"]
      elif type is "2":
         # symmetryKPoints = ["X", "P", "G", "X", "Y", "S", "G", "Z", "S1", "N", "P", "Y1", "Z"]
         symmetryKPoints = [["G","X","Y","S","G","Z","S1","N","P","Y1","Z"],["X","P"]]
         # ["G", "X", "Y", "S", "G", "Z", "S1", "N", "P", "Y1", "Z"], ["X", "P"]
   elif crystalStructure.lower() in ["orthorhombic", "orc", "op"]:
      # 6: ucv = pcv = [[1 0 0]', [0 2 0]', [0 0 3]']
      # symmetryKPoints = ["G","X","U","R","X","S","R","T","S","Y","T","Z","Y","G","Z","U","G"]
      symmetryKPoints = [["G","X","S","Y","G","Z","U","R","T","Z"],["Y","T"],["U","X"],["S","R"]]
      # ["G", "X", "S", "Y", "G", "Z", "U", "R", "T", "Z"], ["Y", "T"], ["U", "X"], ["S", "R"]
   elif crystalStructure.lower() in ["facecenteredorthorhombic", "orcf", "of"]:
      # 7: pcv = [[0 2/2 3/2]', [1./2., 0, 3/2]', [1./2. 2/2 0]']
      #    ucv = [[1 0 0]', [0 2 0]', [0 0 3]']
      pcv4fcc = np.array([[0, 1./2., 1./2.], [1./2., 0, 1./2.], [1./2., 1./2., 0]]).transpose()
      ucv4fcc = np.eye(3)
      pcv = avecStd
      ucv = pcv*(np.linalg.inv(pcv4fcc)*ucv4fcc)
      a = np.norm(ucv[:,0])
      b = np.norm(ucv[:,1])
      c = np.norm(ucv[:,2])
      d = (1./b**2 + 1./c**2 - 1./a**2) * max([a**2, b**2, c**2])
      if  d < -epsilon:
         type = "1"
      elif d > epsilon:
         type = "2"
      else:
         type = "3"
      if type is "1":
         # symmetryKPoints = [["G", "Y", "T", "Z", "G", "X", "A1", "Y"], ["T", "X1", "G", "L"], ["X", "A", "Z"]]
         symmetryKPoints = [["G","Y","T","Z","G","X","A1","Y"],["T","X1"],["X","A","Z"],["L","G"]]
         # ["G", "Y", "T", "Z", "G", "X", "A1", "Y"], ["T", "X1"], ["X", "A", "Z"], ["L", "G"]
      elif type is "2":
         # symmetryKPoints = [["G", "Y", "C", "D", "X", "G", "Z", "D1", "H", "C"], ["X", "H1", "G", "C1", "Z"], ["L", "G", "H", "Y"]]
         symmetryKPoints = [["G","Y","C","D","X","G","Z","D1","H","C"],["C1","Z"],["X","H1"],["H","Y"],["L","G"]]
         # ["G", "Y", "C", "D", "X", "G", "Z", "D1", "H", "C"], ["C1", "Z"], ["X", "H1"], ["H", "Y"], ["L", "G"]
      elif type is "3":
         # symmetryKPoints = [["L", "G", "Y", "T", "Z", "G", "X", "A1", "Y"], ["X", "A", "Z"]]
         symmetryKPoints = [["G","Y","T","Z","G","X","A1","Y"],["X","A","Z"],["L","G"]]
         # ["G", "Y", "T", "Z", "G", "X", "A1", "Y"], ["X", "A", "Z"], ["L", "G"]
   elif crystalStructure.lower() in ["bodycenteredorthorhombic", "orci", "oi"]:
      # 8: pcv = [[-1./2. 2/2 3/2]', [1./2., -2/2, 3/2]', [1./2. 2/2 -3/2]']
      #    ucv = [[1 0 0]', [0 2 0]', [0 0 3]']
      # symmetryKPoints = [["G", "X", "L", "T", "W", "R", "X1", "Z", "G", "Y", "S", "W"], ["Y", "L1", "G", "Y1", "Z"]]
      symmetryKPoints = [["G","X","L","T","W","R","X1","Z","G","Y","S","W"],["L1","Y"],["Y1","Z"]]
      # ["G", "X", "L", "T", "W", "R", "X1", "Z", "G", "Y", "S", "W"], ["L1", "Y"], ["Y1", "Z"]
   elif crystalStructure.lower() in ["ccenteredorthorhombic", "orcc", "os"]:
      # 9: pcv = [[1./2. -2/2 0]', [1./2., 2/2, 0]', [0 0 3]']
      #    ucv = [[1 0 0]', [0 2 0]', [0 0 3]']
      # symmetryKPoints = ["Z","T","G","X","S","R","A","Z","G","Y","X1","A1","T","Y"]
      symmetryKPoints = [["G","X","S","R","A","Z","G","Y","X1","A1","T","Y"],["Z","T"]]
      # ["G", "X", "S", "R", "A", "Z", "G", "Y", "X1", "A1", "T", "Y"], ["Z", "T"]
   elif crystalStructure.lower() in ["hexagonal", "hex", "hcp", "hp"]:
      # 10: ucv = pcv = [[1./2. -sqrt(3)/2 0]', [1./2., sqrt(3)/2, 0]', [0 0 3]']
      # symmetryKPoints = [["G", "M", "K", "G", "A", "L", "H", "A"], ["M", "L", "G", "K", "H"]]
      symmetryKPoints = [["G","M","K","G","A","L","H","A"],["L","M"],["K","H"]]
      # ["G", "M", "K", "G", "A", "L", "H", "A"], ["L", "M"], ["K", "H"]
   elif crystalStructure.lower() in ["rhombohedral", "rhl", "hr"]:
      # 11: ucv = pcv = [[cos(pi/5) -sin(pi/5) 0]', [cos(pi/5) sin(pi/5), 0]',
      #      [cos(2*pi/5)/cos(pi/5), 0, sqrt(1-(cos(2*pi/5)/cos(pi/5))**2)]']
      if np.dot(avecStd[:,2],avecStd[:,0]) > 0:
         type = "1"
      else:
         type = "2"
      if type is "1":
         # symmetryKPoints = [["L","P","G","L","B1","G","B","Z","G","X"],["Q","F","P1","Z"]]
         symmetryKPoints = [["G","L","B1"],["B","Z","G","X"],["Q","F","P1","Z"],["L","P"]]
         # ["G", "L", "B1"], ["B", "Z", "G", "X"], ["Q", "F", "P1", "Z"], ["L", "P"]
      elif type is "2":
         # symmetryKPoints = ["G", "P", "Z", "Q", "G", "F", "P1", "Q1", "L", "Z"]
         symmetryKPoints = [["G","P","Z","Q","G","F","P1","Q1","L","Z"]]
         # ["G", "P", "Z", "Q", "G", "F", "P1", "Q1", "L", "Z"]
   elif crystalStructure.lower() in ["monoclinic", "mcl", "mp"]:
      # 12: ucv = pcv = [[1 0 0]', [0 2 0]', [0, 3*cos(pi/5), 3*sin(pi/5)]']
      # symmetryKPoints = ["Y","D","G","Y","H","G","E","M1","A","X","H1","G","M","D","Z"]
      symmetryKPoints = [["G","Y","H","C","E","M1","A","X","G","Z","D","M"],["Z","A"],["D","Y"],["X","H1"]]
      # ["G", "Y", "H", "G", "E", "M1", "A", "X", "H1"], ["M", "D", "Z"], ["Y", "D"]
   elif crystalStructure.lower() in ["ccenteredmonoclinic", "mclc", "ms"]:
      # 13: pcv = [[1./2. 2/2 0]', [-1./2. 2/2 0]', [0, 3*cos(pi/5), 3*sin(pi/5)]']
      #     ucv = [[1 0 0]', [0 2 0]', [0, 3*cos(pi/5), 3*sin(pi/5)]']
      pcv = avecStd
      a = np.norm(pcv[:,0]-pcv[:,1])
      b = np.norm(pcv[:,0]+pcv[:,1])
      c = np.norm(pcv[:,2])
      kpcv = np.linalg.inv(pcv).transpose()
      kgamma = np.dot(kpcv[:,0],kpcv[:,1])/np.norm(kpcv[:,0])/np.norm(kpcv[:,1])
      if kgamma < -epsilon:
         type = "1"
      elif abs(kgamma) < epsilon:
         type = "2"
      else:
         cosalpha = np.dot(pcv[:,0]+pcv[:,1],pcv[:,2])/b/c
         sin2alpha = 1.-cosalpha**2
         d = b*cosalpha/c + (b**2*sin2alpha/a**2) - 1.
         if d < -epsilon:
            type = "3"
         elif abs(d) < epsilon:
            type = "4"
         else:
            type = "5"
      if type is "1":
         # symmetryKPoints = [["M","G","Y","F","L","I","G","I1","Z","F1","G","X1","Y"],["X", "G", "N"]]
         symmetryKPoints = [["G","Y","F","L","I"],["I1","Z","G","X"],["X1","Y"],["M","G","N"],["Z","F1"]]
         # ["G","Y", "F", "L", "I"], ["I1", "Z", "F1"], ["Y", "X1"], ["X", "G", "N"], ["M", "G"]
      elif type is "2":
         # symmetryKPoints = ["N","G","Y","F","L","I","G","I1","Z","F1","G","M"]
         symmetryKPoints = [["G","Y","F","L","I"],["I1","Z","F1"],["N","G","M"]]
         # ["G", "Y", "F", "L", "I"], ["I1", "Z", "F1"], ["N", "G", "M"]
      elif type is "3":
         # symmetryKPoints = ["M","G","Y","F","H","Z","I","F1","G","H1","Y1","X","G","N"]
         symmetryKPoints = [["G","Y","F","H","Z","I","X","G","Z"],["M","G","N"],["X","Y1","H1"],["I","F1"]]
         # ["G", "Y", "F", "H", "Z", "I", "F1"], ["H1", "Y1", "X", "G", "N"], ["M", "G"]
      elif type is "4":
         # symmetryKPoints = ["M","G","Y","F","H","Z","I","G","H1","Y1","X","G","N"]
         symmetryKPoints = [["G","Y","F","H","Z","I"],["H1","Y1","X","G","N"],["M","G"]]
         # ["G", "Y", "F", "H", "Z", "I"], ["H1", "Y1", "X", "G", "N"], ["M", "G"]
      elif type is "5":
         # symmetryKPoints = ["M","G","Y","F","L","I","G","I1","Z","H","F1","G","H1","Y1","X","G","N"]
         symmetryKPoints = [["G","Y","F","L","I"],["I1","Z","G","X","Y1","H1"],["H","F1"],["F2","X"],["M","G","N"],["H","Z"]]
         # ["G", "Y", "F", "L", "I"], ["I1", "Z", "H", "F1"], ["H1", "Y1", "X", "G", "N"], ["M", "G"]
   elif crystalStructure.lower() in ["triclinic", "tri", "ap"]:
      # 14: ucv = pcv = [[1 0 0]', [2*cos(pi/3), 2*sin(pi/3), 0]', [3*cos(pi/4),
      #      3/sin(pi/3)*(cos(pi/5)-cos(pi/4)*cos(pi/3)),
      #      3/sin(pi/3)*sqrt(1-(cos(pi/5))**2-(cos(pi/4))**2-(cos(pi/3))**2
      #      + 2*cos(pi/5)*cos(pi/4)*cos(pi/3))]'] : [1 2 3], [pi/5,pi/4,pi/3]
      kpcv = np.linalg.inv(avecStd).transpose()
      a = np.norm(kpcv[:,0])
      b = np.norm(kpcv[:,1])
      c = np.norm(kpcv[:,2])
      kalpha = np.dot(kpcv[:,1],kpcv[:,2])/b/c
      kbeta = np.dot(kpcv[:,2],kpcv[:,0])/c/a
      kgamma = np.dot(kpcv[:,0],kpcv[:,1])/a/b
      if kalpha < 0:
         if kgamma < -epsilon:
            type = "1a"
         else:
            type = "2a"
      else:
         if kgamma > epsilon:
            type = "1b"
         else:
            type = "2b"
      if type in ["1a","2a","1b","2b"]:
         symmetryKPoints = [["X", "G", "Y"], ["L", "G", "Z"], ["N", "G", "M"], ["R", "G"]]
         # ["X", "G", "Y"], ["L", "G", "Z"], ["N", "G", "M"], ["R", "G"]
   elif type is "0D":
      # 0D: isolate : [10 10 10]
      symmetryKPoints = ["G"]
   elif type is "1D":
      # 1D: along z direction. c << a,b : [10 10 1]
      symmetryKPoints = ["G", "X"]
   elif crystalStructure.lower() in ["square2d", "squ2d"]:
      # 2D1: along x-y plane. c >> a,b
      #     [[1 0 0]', [0, 1, 0]', [0 0 10]'] : [1 1 10]
      #     analog: Tetragonal
      symmetryKPoints = ["G", "X", "M", "G"]
   elif crystalStructure.lower() in ["rectangular2d", "rectangle2d", "rec2d"]:
      # 2D2: along x-y plane. c >> a,b
      #     [[1 0 0]', [0, 2, 0]', [0 0 10]'] : [1 1 10]
      #     analog: Orthorhombic
      symmetryKPoints = ["G", "X", "S", "Y", "G", "S"]
   elif crystalStructure.lower() in ["hexagonal2d", "hexagon2d", "hex2d"]:
      # 2D3: along x-y plane. c >> a,b
      #     [[1./2. -sqrt(3)/2 0]', [1./2., sqrt(3)/2, 0]', [0 0 10]'] : [1 1 10]
      #     analog: Hexagonal
      symmetryKPoints = ["G", "M", "K", "G"]
   elif crystalStructure.lower() in ["rhombic2d", "rhombus2d", "centeredrectangular2d", "rho2d"]:
      # 2D4:  along x-y plane. c >> a,b
      #     [[1./2. -2/2 0]', [1./2., 2/2, 0]', [0 0 10]'] : [1 2 10]
      #     analog: CCenteredOrthorhombic
      #     pcv = [[1./2. -2/2 0]', [1./2., 2/2, 0]', [0 0 10]']
      #     ucv = [[1 0 0]', [0 2 0]', [0 0 10]']
      symmetryKPoints = ["S","G","X","S","X1","Y","G","X1"]
   elif crystalStructure.lower() in ["oblique2d", "obl2d"]:
      # 2D5:  along x-y plane. c >> a,b
      #     [[2 0 0]', [3*cos(pi/5), 3*sin(pi/5), 0]', [0 0 10]'] : [2 3 10]
      #     analog: "Monoclinic"
      symmetryKPoints = ["H","G","Y","H","C","G","X","H1","G","Y1","H2"]
   else:
      raise NameError("Crystal structure not recognized.")
   return symmetryKPoints

def kValue2kSymbol(crystalStructure, primitiveCellVectors, kValue):
   """
   This function finds a high-symmetry k-point label from reduced coordinates.

   Parameters
   ----------
   crystalStructure : string
      Name of the Bravais lattice.

   primitiveCellVectors : 2D array
      3 x 3 array of row-vectors corresponding to the lattice vectors).

   kValue : 1D array
      High-symmetry k-point fractional coordinates.

   Returns
   -------
   kSymbol: string
      High-symmetry k-point label.

   Author Info:
   Vincent Michaud-Rioux
   2020-06-23
   """
   if isinstance(kValue, list):
      return [kValue2kSymbol(crystalStructure, primitiveCellVectors, k) for k in kValue]

   if isinstance(kValue, str):
      return kValue

   kptDict = kSymbol_2_kValue_dict(crystalStructure, primitiveCellVectors)
   kSymbol = "!"
   for i in kptDict.keys():
      if np.allclose(kValue, kptDict[i], rtol=1.e-5, atol=1.e-5):
         kSymbol = i
         break
   return kSymbol

def kSymbol2kValue(crystalStructure, primitiveCellVectors, kSymbol):
   """
   This function is to convert a high-symmetry k-point label to reduced coordinates.

   This function is to convert a conventional symbol for a high
   symmetrical point in k space into its corresponding coordinate
   value, where the unit of the value is the three reciprocal primitive
   cell vectors.
   Please note that each symbol actually represents a set of infinitely 
   many equivalent points and the output coordinates is just for one of them.

   Parameters
   ----------
   crystalStructure : string
      Name of the Bravais lattice.

   primitiveCellVectors : 2D array
      3 x 3 array of row-vectors corresponding to the lattice vectors).

   kSymbol: string
      High-symmetry k-point label.

   Returns
   -------
   kValue : 1D array
      High-symmetry k-point fractional coordinates.

   Author Info:
   Vincent Michaud-Rioux
   2020-06-23
   """
   if isinstance(kSymbol, list):
      return [kSymbol2kValue(crystalStructure, primitiveCellVectors, k) for k in kSymbol]

   if not isinstance(kSymbol, str):
      return np.array(kSymbol)

   kptDict = kSymbol_2_kValue_dict(crystalStructure, primitiveCellVectors)
   if not kSymbol in kptDict.keys():
      NameError("Unknown symbol "+kSymbol+" for Bravais lattice "+crystalStructure+".")        
   return np.array(kptDict[kSymbol])

def kSymbol_2_kValue_dict(crystalStructure, primitiveCellVectors):
   """
   This function returns a dictionary of the high-symmetry k-point labels
   of a given crystal structure in reduced coordinates.

   Parameters
   ----------
   crystalStructure : string
      Name of the Bravais lattice.

   primitiveCellVectors : 2D array
      3 x 3 array of row-vectors corresponding to the lattice vectors).

   Returns
   -------
   kptDict : dict
      High-symmetry k-points in fractional coordinates.

   Author Info:
   Lei Liu, HZWTech
   2015-10
   Vincent Michaud-Rioux
   2020-06-23

   ############## about the symbol of the symmetry k points ##################
   ## This is simply the convention for labeling points of high symmetry in
   ## the Brillouin zone.  It"s origin is: Bouckaert, Smoluchowski, and
   ## Wigner, "Theory of Brillouin Zones and Symmetry Properties of Wave
   ## Functions in Crystals," Phys. Rev. 50, p. 58 (1936).  The guide to all
   ## the details of this is John C. Slater, "Symmetry and Energy Bands in
   ## Crystals" (reprinted by Dover in 1972).  It has all the 3-d crystal
   ## structures and, I suspect, 2-d as well.  These labels may make some
   ## sense in the context of group theory (or perhaps in Hungarian), but
   ## they are commonly regarded as an arbitrary but common convention.
   ##
   ## It is also called critical point. see http://en.wikipedia.org/wiki/Brillouin_zone
   ###########################################################################
   """

   primitiveCellVectors = primitiveCellVectors.transpose() # here primitiveCellVectors = [a1 a2 a3]
   kptDict = {}
   epsilon = 1e-3
   if crystalStructure.lower() in ["cubic","cub","simplecubic","sc","cp"]:
      # 1: ucv = pcv = [[1 0., 0]',[0., 1 0]',[0., 0., 1]']
      kSymbols = ["G","X","M","R"]
      kptDict[kSymbols[0]] = [0., 0., 0.] # Gamma: Center of the Brillouin zone
      kptDict[kSymbols[1]] = [0., .5, 0.] # X: Center of a face
      kptDict[kSymbols[2]] = [.5, .5, 0.] # M: Center of an edge
      kptDict[kSymbols[3]] = [.5, .5, .5] # R: Corner point
   elif crystalStructure.lower() in ["facecenteredcubic","fcc","cf"]:
      # 2: pcv = [[0., .5, .5]',[.5,0,.5]',[.5, .5, 0]']
      #    ucv = [[1 0., 0]',[0., 1 0]',[0., 0., 1]']
      kSymbols = ["G","K","L","U","W","X"]
      kptDict[kSymbols[0]] = [0., 0., 0]  # Gamma: Center of the Brillouin zone
      kptDict[kSymbols[1]] = [3/8, 3/8, 3/4] # K: Middle of an edge joining two hexagonal faces
      kptDict[kSymbols[2]] = [.5, .5, .5] # L: Center of a hexagonal face
      kptDict[kSymbols[3]] = [5/8, 1/4, 5/8] # U: Middle of an edge joining a hexagonal and a square face
      kptDict[kSymbols[4]] = [.5, 1/4, 3/4]  # W : Corner point
      kptDict[kSymbols[5]] = [.5, 0., .5]  # X: Center of a square face
   elif crystalStructure.lower() in ["bodycenteredcubic","bcc","ci"]:
      # 3: pcv = [[-.5, .5, .5]',[.5,-.5,.5]',[.5, .5, -.5]']
      #    ucv = [[1 0., 0]',[0., 1 0]',[0., 0., 1]']
      kSymbols = ["G","H","P","N"]
      kptDict[kSymbols[0]] = [0., 0., 0]  # Gamma: Center of the Brillouin zone
      kptDict[kSymbols[1]] = [.5, -.5, .5] # H: Corner point joining four edges
      kptDict[kSymbols[2]] = [1/4, 1/4, 1/4] # P: Corner point joining three edges
      kptDict[kSymbols[3]] = [0., 0., .5] # N: Center of a face
   elif crystalStructure.lower() in ["tetragonal","tet","tp"]:
      # 4: ucv = pcv = [[1 0., 0]',[0., 1 0]',[0., 0., 5]']
      kSymbols = ["G","A","M","R","X","Z"]
      kptDict[kSymbols[0]] = [0., 0., 0]  # Gamma: Center of the Brillouin zone
      kptDict[kSymbols[1]] = [.5, .5, .5]
      kptDict[kSymbols[2]] = [.5, .5, 0]
      kptDict[kSymbols[3]] = [0., .5, .5]
      kptDict[kSymbols[4]] = [0., .5, 0]
      kptDict[kSymbols[5]] = [0., 0., .5]
   elif crystalStructure.lower() in ["bodycenteredtetragonal","bct","ti"]:
      # 5: pcv = [[-.5, .5, 5/2]',[.5,-.5,5/2]',[.5, .5, -5/2]']
      #    ucv = [[1 0., 0]',[0., 1 0]',[0., 0., 5]']
      pcv4bcc = np.array([[-.5, .5, .5],[.5,-.5,.5],[.5, .5, -.5]]).transpose()
      ucv4bcc = np.eye(3)
      pcv = primitiveCellVectors
      ucv = np.matmul(pcv, np.linalg.solve(pcv4bcc,ucv4bcc))
      if np.linalg.norm(ucv[:,2]) < np.linalg.norm(ucv[:,0]):
         type = "1"
      else:
         type = "2"
      if type == "1":
         a = np.linalg.norm(ucv[:,0])
         c = np.linalg.norm(ucv[:,2])
         eta = (1 + c**2/a**2)/4
         kSymbols = ["G","M","N","P","X","Z","Z1"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [-.5, .5, .5]
         kptDict[kSymbols[2]] = [0., .5, 0]
         kptDict[kSymbols[3]] = [1/4, 1/4, 1/4]
         kptDict[kSymbols[4]] = [0., 0., .5]
         kptDict[kSymbols[5]] = [eta, eta, -eta]
         kptDict[kSymbols[6]] = [-eta, 1-eta, eta]
      elif type == "2":
         a = np.linalg.norm(ucv[:,0])
         c = np.linalg.norm(ucv[:,2])
         eta = (1. + a**2/c**2)/4.
         zeta = a**2/c**2/2.
         kSymbols = ["G","N","P","X","Z","S","S1","Y","Y1"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [0., .5, 0]
         kptDict[kSymbols[2]] = [1/4, 1/4, 1/4]
         kptDict[kSymbols[3]] = [0., 0., .5]
         kptDict[kSymbols[4]] = [.5, .5, -.5]
         kptDict[kSymbols[5]] = [-eta, eta, eta]
         kptDict[kSymbols[6]] = [eta, 1-eta, -eta]
         kptDict[kSymbols[7]] = [-zeta, zeta, .5]
         kptDict[kSymbols[8]] = [.5, .5, -zeta]
   elif crystalStructure.lower() in ["orthorhombic","orc","op"]:
      # 6: ucv = pcv = [[1 0., 0]',[0., 2 0]',[0., 0., 3]']
      kSymbols = ["G","R","S","T","U","X","Y","Z"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [.5, .5, .5]
      kptDict[kSymbols[2]] = [.5, .5, 0]
      kptDict[kSymbols[3]] = [0., .5, .5]
      kptDict[kSymbols[4]] = [.5, 0., .5]
      kptDict[kSymbols[5]] = [.5, 0., 0]
      kptDict[kSymbols[6]] = [0., .5, 0]
      kptDict[kSymbols[7]] = [0., 0., .5]
   elif crystalStructure.lower() in ["facecenteredorthorhombic","orcf","of"]:
      # 7: pcv = [[0., 2/2 3/2]',[.5,0,3/2]',[.5, 2/2 0]']
      #    ucv = [[1 0., 0]',[0., 2 0]',[0., 0., 3]']
      pcv4fcc = np.array([[0., .5, .5],[.5,0,.5],[.5, .5, 0]]).transpose()
      ucv4fcc = np.eye(3)
      pcv = primitiveCellVectors
      ucv = np.matmul(pcv, np.linalg.solve(pcv4fcc, ucv4fcc))
      a = np.linalg.norm(ucv[:,0])
      b = np.linalg.norm(ucv[:,1])
      c = np.linalg.norm(ucv[:,2])
      d = (1/b**2 + 1/c**2 - 1/a**2) * max([a**2,b**2,c**2])
      if  d < -epsilon:
         type = "1"
      elif d > epsilon:
         type = "2"
      else:
         type = "3"
      if type in ["1", "3"]:
         eta = (1. + a**2/b**2 + a**2/c**2)/4.
         zeta = (1. + a**2/b**2 - a**2/c**2)/4.
         kSymbols = ["G","L","T","Y","Z","A","A1","X","X1"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, .5, .5]
         kptDict[kSymbols[2]] = [1, .5, .5]
         kptDict[kSymbols[3]] = [.5, 0., .5]
         kptDict[kSymbols[4]] = [.5, .5, 0]
         kptDict[kSymbols[5]] = [.5, .5+zeta, zeta]
         kptDict[kSymbols[6]] = [.5, .5-zeta, 1-zeta]
         kptDict[kSymbols[7]] = [0., eta, eta]
         kptDict[kSymbols[8]] = [1, 1-eta, 1-eta]
      elif type == "2":
         eta = (1 + a**2/b**2 - a**2/c**2)/4
         delta = (1 + b**2/a**2 - b**2/c**2)/4
         phi = (1 + c**2/b**2 - c**2/a**2)/4
         kSymbols = ["G","L","X","Y","Z","C","C1","D","D1","H","H1"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, .5, .5]
         kptDict[kSymbols[2]] = [0., .5, .5]
         kptDict[kSymbols[3]] = [.5, 0., .5]
         kptDict[kSymbols[4]] = [.5, .5, 0]
         kptDict[kSymbols[5]] = [.5, .5-eta, 1-eta]
         kptDict[kSymbols[6]] = [.5, .5+eta, eta]
         kptDict[kSymbols[7]] = [.5-delta, .5, 1-delta]
         kptDict[kSymbols[8]] = [.5+delta, .5, delta]
         kptDict[kSymbols[9]] = [1-phi, .5-phi, .5]
         kptDict[kSymbols[10]] = [phi, .5+phi, .5]
   elif crystalStructure.lower() in ["bodycenteredorthorhombic","orci","oi"]:
      # 8: pcv = [[-.5, 2/2 3/2]',[.5,-2/2,3/2]',[.5, 2/2 -3/2]']
      #    ucv = [[1 0., 0]',[0., 2 0]',[0., 0., 3]']
      pcv4bcc = np.array([[-.5, .5, .5],[.5,-.5,.5],[.5, .5, -.5]]).transpose()
      ucv4bcc = np.eye(3)
      pcv = primitiveCellVectors
      ucv = np.matmul(pcv, np.linalg.solve(pcv4bcc, ucv4bcc))
      a = np.linalg.norm(ucv[:,0])
      b = np.linalg.norm(ucv[:,1])
      c = np.linalg.norm(ucv[:,2])
      zeta = (1 + a**2/c**2)/4
      eta = (1 + b**2/c**2)/4
      delta = (b**2 - a**2)/(4*c**2)
      mu = (a**2 + b**2)/(4*c**2)
      kSymbols = ["G","R","S","T","W","Z","L","L1","L2","X","X1","Y","Y1"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [0., .5, 0]
      kptDict[kSymbols[2]] = [.5, 0., 0]
      kptDict[kSymbols[3]] = [0., 0., .5]
      kptDict[kSymbols[4]] = [1/4, 1/4, 1/4]
      kptDict[kSymbols[5]] = [.5, .5, -.5]
      kptDict[kSymbols[6]] = [-mu, mu, .5-delta]
      kptDict[kSymbols[7]] = [mu, -mu, .5+delta]
      kptDict[kSymbols[8]] = [.5-delta, .5+delta, -mu]
      kptDict[kSymbols[9]] = [-zeta, zeta, zeta]
      kptDict[kSymbols[10]] = [zeta, 1-zeta, -zeta]
      kptDict[kSymbols[11]] = [eta, -eta, eta]
      kptDict[kSymbols[12]] = [1-eta, eta, -eta]
   elif crystalStructure.lower() in ["ccenteredorthorhombic","orcc","os"]:
      # 9: pcv = [[.5, -2/2 0]',[.5,2/2,0]',[0., 0., 3]']
      #    ucv = [[1 0., 0]',[0., 2 0]',[0., 0., 3]']
      pcv = primitiveCellVectors
      a = np.linalg.norm(pcv[:,0]+pcv[:,1])
      b = np.linalg.norm(pcv[:,0]-pcv[:,1])
      zeta = (1 + a**2/b**2)/4
      kSymbols = ["G","R","S","T","Y","Z","A","A1","X","X1"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [0., .5, .5]
      kptDict[kSymbols[2]] = [0., .5, 0]
      kptDict[kSymbols[3]] = [-.5, .5, .5]
      kptDict[kSymbols[4]] = [-.5, .5, 0]
      kptDict[kSymbols[5]] = [0., 0., .5]
      kptDict[kSymbols[6]] = [zeta, zeta, .5]
      kptDict[kSymbols[7]] = [-zeta, 1-zeta, .5]
      kptDict[kSymbols[8]] = [zeta, zeta, 0]
      kptDict[kSymbols[9]] = [-zeta, 1-zeta, 0]
   elif crystalStructure.lower() in ["hexagonal","hex","hcp","hp"]:
      # 10: ucv = pcv = [[.5, -sqrt(3)/2 0]',[.5,sqrt(3)/2,0]',[0., 0., 3]']
      kSymbols = ["G","A","H","K","L","M"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [0., 0., .5]
      kptDict[kSymbols[2]] = [1/3, 1/3, .5]
      kptDict[kSymbols[3]] = [1/3, 1/3, 0]
      kptDict[kSymbols[4]] = [.5, 0., .5]
      kptDict[kSymbols[5]] = [.5, 0., 0]
   elif crystalStructure.lower() in ["rhombohedral","rhl","hr"]:
      # 11: ucv = pcv = [[cos(pi/5) -sin(pi/5) 0]',[cos(pi/5) sin(pi/5),0]',
      #      [cos(2*pi/5)/cos(pi/5),0,sqrt(1-(cos(2*pi/5)/cos(pi/5))**2)]']
      pcv = primitiveCellVectors
      cosalpha = np.dot(pcv[:,2],pcv[:,0])/np.linalg.norm(pcv[:,0])/np.linalg.norm(pcv[:,2])
      if cosalpha > 0:
         type = "1"
      else:
         type = "2"
      if type == "1":
         eta = (1 + 4*cosalpha)/(2 + 4*cosalpha)
         nu = 3/4 - eta/2
         kSymbols = ["G","F","L","L1","Z","B","B1","P","P1","P2","Q","X"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, .5, 0]
         kptDict[kSymbols[2]] = [.5, 0., 0]
         kptDict[kSymbols[3]] = [0., 0., -.5]
         kptDict[kSymbols[4]] = [.5, .5, .5]
         kptDict[kSymbols[5]] = [eta, .5, 1-eta]
         kptDict[kSymbols[6]] = [.5, 1-eta, eta-1]
         kptDict[kSymbols[7]] = [eta, nu, nu]
         kptDict[kSymbols[8]] = [1-nu, 1-nu, 1-eta]
         kptDict[kSymbols[9]] = [nu, nu, eta-1]
         kptDict[kSymbols[10]] = [1-nu, nu, 0]
         kptDict[kSymbols[11]] = [nu, 0., -nu]
      elif type == "2":
         eta = (.5)*(1+cosalpha)/(1-cosalpha)
         nu = 3/4 - eta/2
         kSymbols = ["G","F","L","Z","P","P1","Q","Q1"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, -.5, 0]
         kptDict[kSymbols[2]] = [.5, 0., 0]
         kptDict[kSymbols[3]] = [.5, -.5, .5]
         kptDict[kSymbols[4]] = [1-nu, -nu, 1-nu]
         kptDict[kSymbols[5]] = [nu, nu-1, nu-1]
         kptDict[kSymbols[6]] = [eta, eta, eta]
         kptDict[kSymbols[7]] = [1-eta, -eta, -eta]
   elif crystalStructure.lower() in ["monoclinic","mcl","mp"]:
      # 12: ucv = pcv = [[1 0., 0]',[0., 2 0]',[0,3*cos(pi/5),3*sin(pi/5)]']
      pcv = primitiveCellVectors
      b = np.linalg.norm(pcv[:,1])
      c = np.linalg.norm(pcv[:,2])
      cosalpha = np.dot(pcv[:,1],pcv[:,2])/b/c
      sin2alpha = 1-cosalpha**2
      eta = (1 - b*cosalpha/c)/(2*sin2alpha)
      nu = .5 - eta*c*cosalpha/b
      kSymbols = ["G","A","C","D","D1","E","X","Y","Y1","Z","H","H1","H2","M","M1","M2"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [.5, .5, 0]
      kptDict[kSymbols[2]] = [0., .5, .5]
      kptDict[kSymbols[3]] = [.5, 0., .5]
      kptDict[kSymbols[4]] = [.5, 0., -.5]
      kptDict[kSymbols[5]] = [.5, .5, .5]
      kptDict[kSymbols[6]] = [0., .5, 0]
      kptDict[kSymbols[7]] = [0., 0., .5]
      kptDict[kSymbols[8]] = [0., 0., -.5]
      kptDict[kSymbols[9]] = [.5, 0., 0]
      kptDict[kSymbols[10]] = [0., eta, 1-nu]
      kptDict[kSymbols[11]] = [0., 1-eta, nu]
      kptDict[kSymbols[12]] = [0., eta, -nu]
      kptDict[kSymbols[13]] = [.5, eta, 1-nu]
      kptDict[kSymbols[14]] = [.5, 1-eta, nu]
      kptDict[kSymbols[15]] = [.5, eta, -nu]
   elif crystalStructure.lower() in ["ccenteredmonoclinic","mclc","ms"]:
      # 13: pcv = [[.5, 2/2 0]',[-.5, 2/2 0]',[0,3*cos(pi/5),3*sin(pi/5)]']
      #     ucv = [[1 0., 0]',[0., 2 0]',[0,3*cos(pi/5),3*sin(pi/5)]']
      pcv = primitiveCellVectors
      a = np.linalg.norm(pcv[:,0]-pcv[:,1])
      b = np.linalg.norm(pcv[:,0]+pcv[:,1])
      c = np.linalg.norm(pcv[:,2])
      kpcv = np.linalg.inv(pcv).transpose()
      kgamma = np.dot(kpcv[:,0],kpcv[:,1])/np.linalg.norm(kpcv[:,0])/np.linalg.norm(kpcv[:,1])
      if kgamma < -epsilon:
         type = "1"
      elif abs(kgamma) < epsilon:
         type = "2"
      else:
         cosalpha = np.dot(pcv[:,0]+pcv[:,1],pcv[:,2])/b/c
         sin2alpha = 1-cosalpha**2
         d = b*cosalpha/c + b**2*sin2alpha/a**2 - 1
         if d < -epsilon:
            type = "3"
         elif abs(d) < epsilon:
            type = "4"
         else:
            type = "5"
      if type in ["1", "2"]:
         cosalpha = np.dot(pcv[:,0]+pcv[:,1],pcv[:,2])/b/c
         sin2alpha = 1-cosalpha**2
         zeta = (2-b*cosalpha/c)/(4*sin2alpha)
         eta = .5 + 2*zeta*c*cosalpha/b
         psi = 3/4 - a**2/(4*b**2*sin2alpha)
         phi = psi + (3/4-psi)*b*cosalpha/c
         kSymbols = ["G","N","N1","L","M","Y","Y1","Z","F","F1","F2","F3","I","I1","X","X1","X2"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, 0., 0]
         kptDict[kSymbols[2]] = [0., -.5, 0]
         kptDict[kSymbols[3]] = [.5, .5, .5]
         kptDict[kSymbols[4]] = [.5, 0., .5]
         kptDict[kSymbols[5]] = [.5, .5, 0]
         kptDict[kSymbols[6]] = [-.5, -.5, 0]
         kptDict[kSymbols[7]] = [0., 0., .5]
         kptDict[kSymbols[8]] = [1-zeta, 1-zeta, 1-eta]
         kptDict[kSymbols[9]] = [zeta, zeta, eta]
         kptDict[kSymbols[10]] = [-zeta, -zeta, 1-eta]
         kptDict[kSymbols[11]] = [1-zeta, -zeta, 1-eta]
         kptDict[kSymbols[12]] = [phi, 1-phi, .5]
         kptDict[kSymbols[13]] = [1-phi, phi-1, .5]
         kptDict[kSymbols[14]] = [1-psi, psi-1, 0]
         kptDict[kSymbols[15]] = [psi, 1-psi, 0]
         kptDict[kSymbols[16]] = [psi-1, -psi, 0]
      elif type in ["3", "4"]:
         cosalpha = np.dot(pcv[:,0]+pcv[:,1],pcv[:,2])/b/c
         sin2alpha = 1-cosalpha**2
         mu = (1 + b**2/a**2)/4
         delta = b*c*cosalpha/(2*a**2)
         zeta = mu - 1/4 + (1-b*cosalpha/c)/(4*sin2alpha)
         eta = .5 + 2*zeta*c*cosalpha/b
         phi = 1 + zeta - 2*mu
         psi = eta - 2*delta
         kSymbols = ["G","I","M","N","N1","X","Z","F","F1","F2","H","H1","H2","Y","Y1","Y2","Y3"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, -.5, .5]
         kptDict[kSymbols[2]] = [.5, 0., .5]
         kptDict[kSymbols[3]] = [.5, 0., 0]
         kptDict[kSymbols[4]] = [0., -.5, 0]
         kptDict[kSymbols[5]] = [.5, -.5, 0]
         kptDict[kSymbols[6]] = [0., 0., .5]
         kptDict[kSymbols[7]] = [1-phi, 1-phi, 1-psi]
         kptDict[kSymbols[8]] = [phi, phi-1, psi]
         kptDict[kSymbols[9]] = [1-phi, -phi, 1-psi]
         kptDict[kSymbols[10]] = [zeta, zeta, eta]
         kptDict[kSymbols[11]] = [1-zeta, -zeta, 1-eta]
         kptDict[kSymbols[12]] = [-zeta, -zeta, 1-eta]
         kptDict[kSymbols[13]] = [mu, mu, delta]
         kptDict[kSymbols[14]] = [1-mu, -mu, -delta]
         kptDict[kSymbols[15]] = [-mu, -mu, -delta]
         kptDict[kSymbols[16]] = [mu, mu-1, delta]
      elif type == "5":
         cosalpha = np.dot(pcv[:,0]+pcv[:,1],pcv[:,2])/b/c
         sin2alpha = 1-cosalpha**2
         zeta = (b**2/a**2 + (1 - b*cosalpha/c)/sin2alpha)/4
         eta = .5 + 2*zeta*c*cosalpha/b
         mu = eta/2 + b**2/(4*a**2) - b*c*cosalpha/(2*a**2)
         nu = 2*mu - zeta
         omega = (4*nu - 1 - b**2*sin2alpha/a**2)*c/(2*b*cosalpha)
         delta = zeta*c*cosalpha/b + omega/2 - 1/4
         rho = 1 - zeta*a**2/b**2
         kSymbols = ["G","L","M","N","N1","X","Z","F","F1","F2","H","H1","H2","I","I1","Y","Y1","Y2","Y3"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, .5, .5]
         kptDict[kSymbols[2]] = [.5, 0., .5]
         kptDict[kSymbols[3]] = [.5, 0., 0]
         kptDict[kSymbols[4]] = [0., -.5, 0]
         kptDict[kSymbols[5]] = [.5, -.5, 0]
         kptDict[kSymbols[6]] = [0., 0., .5]
         kptDict[kSymbols[7]] = [nu, nu, omega]
         kptDict[kSymbols[8]] = [1-nu, 1-nu, 1-omega]
         kptDict[kSymbols[9]] = [nu, nu-1, omega]
         kptDict[kSymbols[10]] = [zeta, zeta, eta]
         kptDict[kSymbols[11]] = [1-zeta, -zeta, 1-eta]
         kptDict[kSymbols[12]] = [-zeta, -zeta, 1-eta]
         kptDict[kSymbols[13]] = [rho, 1-rho, .5]
         kptDict[kSymbols[14]] = [1-rho, rho-1, .5]
         kptDict[kSymbols[15]] = [mu, mu, delta]
         kptDict[kSymbols[16]] = [1-mu, -mu, -delta]
         kptDict[kSymbols[17]] = [-mu, -mu, -delta]
         kptDict[kSymbols[18]] = [mu, mu-1, delta]
   elif crystalStructure.lower() in ["triclinic","tri","ap"]:
      # 14: ucv = pcv = [[1 0., 0]',[2*cos(pi/3),2*sin(pi/3),0]',[3*cos(pi/4),
      #      3/sin(pi/3)*(cos(pi/5)-cos(pi/4)*cos(pi/3)),
      #      3/sin(pi/3)*sqrt(1-(cos(pi/5))**2-(cos(pi/4))**2-(cos(pi/3))**2
      #      + 2*cos(pi/5)*cos(pi/4)*cos(pi/3))]'] :: [1 2 3],[pi/5,pi/4,pi/3]
      kpcv = np.linalg.inv(primitiveCellVectors).transpose()
      a = np.linalg.norm(kpcv[:,0])
      b = np.linalg.norm(kpcv[:,1])
      c = np.linalg.norm(kpcv[:,2])
      kalpha = np.dot(kpcv[:,1],kpcv[:,2])/b/c
      # kbeta = np.dot(kpcv[:,2],kpcv[:,0])/c/a
      kgamma = np.dot(kpcv[:,0],kpcv[:,1])/a/b
      if kalpha < 0:
         if kgamma < -epsilon:
            type = "1a"
         else:
            type = "2a"
      else:
         if kgamma > epsilon:
            type = "1b"
         else:
            type = "2b"
      if type in ["1a","2a"]:
         kSymbols = ["G","L","M","N","R","X","Y","Z"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, .5, 0]
         kptDict[kSymbols[2]] = [0., .5, .5]
         kptDict[kSymbols[3]] = [.5, 0., .5]
         kptDict[kSymbols[4]] = [.5, .5, .5]
         kptDict[kSymbols[5]] = [.5, 0., 0]
         kptDict[kSymbols[6]] = [0., .5, 0]
         kptDict[kSymbols[7]] = [0., 0., .5]
      elif type in ["1b","2b"]:
         kSymbols = ["G","L","M","N","R","X","Y","Z"]
         kptDict[kSymbols[0]] = [0., 0., 0]
         kptDict[kSymbols[1]] = [.5, -.5, 0]
         kptDict[kSymbols[2]] = [0., 0., .5]
         kptDict[kSymbols[3]] = [-.5, -.5, .5]
         kptDict[kSymbols[4]] = [0., -.5, .5]
         kptDict[kSymbols[5]] = [0., -.5, 0]
         kptDict[kSymbols[6]] = [.5, 0., 0]
         kptDict[kSymbols[7]] = [-.5, 0., .5]
   elif crystalStructure == "0D":
      # 0D: isolate :: [10., 10., 10]
      kSymbols = ["G"]
      kptDict[kSymbols[0]] = [0., 0., 0]
   elif crystalStructure == "1D":
      # 1D: along z direction. c << a,b :: [10., 10., 1]
      kSymbols = ["G","X"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [0., 0., .5]
   elif crystalStructure.lower() in ["square2d","squ2d"]:
      # 2D1: along x-y plane. c >> a,b
      #     [[1 0., 0]',[0,1,0]',[0., 0., 10]'] :: [1 1 10]
      #     analog: Tetragonal
      kSymbols = ["G","X","M"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [0., .5, 0]
      kptDict[kSymbols[2]] = [.5, .5, 0]
   elif crystalStructure.lower() in ["rectangular2d","rectangle2d","rec2d"]:
      # 2D2: along x-y plane. c >> a,b
      #     [[1 0., 0]',[0,2,0]',[0., 0., 10]'] :: [1 1 10]
      #     analog: Orthorhombic
      kSymbols = ["G","X","Y","S"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [.5, 0., 0]
      kptDict[kSymbols[2]] = [0., .5, 0]
      kptDict[kSymbols[3]] = [.5, .5, 0]
   elif crystalStructure.lower() in ["hexagonal2d","hexagon2d","hex2d"]:
      # 2D3: along x-y plane. c >> a,b
      #     [[.5, -sqrt(3)/2 0]',[.5,sqrt(3)/2,0]',[0., 0., 10]'] :: [1 1 10]
      #     analog: Hexagonal
      kSymbols = ["G","M","K"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [.5, 0., 0]
      kptDict[kSymbols[2]] = [1/3, 1/3, 0]
   elif crystalStructure.lower() in ["rhombic2d","rhombus2d","centeredrectangular2d","rho2d"]:
      # 2D4:  along x-y plane. c >> a,b
      #     [[.5, -2/2 0]',[.5,2/2,0]',[0., 0., 10]'] :: [1 2 10]
      #     analog: CCenteredOrthorhombic
      #     pcv = [[.5, -2/2 0]',[.5,2/2,0]',[0., 0., 10]']
      #     ucv = [[1 0., 0]',[0., 2 0]',[0., 0., 10]']
      pcv = primitiveCellVectors
      a = np.linalg.norm(pcv[:,0]+pcv[:,1])
      b = np.linalg.norm(pcv[:,0]-pcv[:,1])
      zeta = (1 + a**2/b**2)/4
      kSymbols = ["G","S","Y","X","X1"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [0., .5, 0]
      kptDict[kSymbols[2]] = [-.5, .5, 0]
      kptDict[kSymbols[3]] = [zeta, zeta, 0]
      kptDict[kSymbols[4]] = [-zeta, 1-zeta, 0]
   elif crystalStructure.lower() in ["oblique2d","obl2d"]:
      # 2D5:  along x-y plane. c >> a,b
      #     [[2 0., 0]',[3*cos(pi/5),3*sin(pi/5),0]',[0., 0., 10]'] :: [2 3 10]
      #     analog: "Monoclinic"
      pcv = primitiveCellVectors
      b = np.linalg.norm(pcv[:,0])
      c = np.linalg.norm(pcv[:,1])
      cosalpha = np.dot(pcv[:,0],pcv[:,1])/b/c
      sin2alpha = 1-cosalpha**2
      eta = (1 - b*cosalpha/c)/(2*sin2alpha)
      nu = .5 - eta*c*cosalpha/b
      kSymbols = ["G","C","X","Y","Y1","H","H1","H2"]
      kptDict[kSymbols[0]] = [0., 0., 0]
      kptDict[kSymbols[1]] = [.5, .5, 0]
      kptDict[kSymbols[2]] = [.5, 0., 0]
      kptDict[kSymbols[3]] = [0., .5, 0]
      kptDict[kSymbols[4]] = [0., -.5, 0]
      kptDict[kSymbols[5]] = [eta, 1-nu, 0]
      kptDict[kSymbols[6]] = [1-eta, nu, 0]
      kptDict[kSymbols[7]] = [eta, -nu, 0]
   else:
      NameError("Unknown crystal structure "+crystalStructure+".")
   return kptDict

###########
# testing #
###########

def print_kvalues(crystalStructure, avec, kSymbols):
   for i in range(0, len(kSymbols)):
      kval = kSymbol2kValue(crystalStructure, avec, kSymbols[i])
      print('%6s %6s %f %f %f' % (crystalStructure, kSymbols[i], kval[0], kval[1], kval[2]))

def print_all_bra2avec():
   perm = [[0,1,2],[0,2,1],[1,0,2],[1,2,0],[2,0,1],[2,1,0]]

   crystalStructure = 'CUB'; avec = bra2avec('CUB')
   kSymbols = ['G','X','M','R']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'CUB')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'FCC'; avec = bra2avec('FCC')
   kSymbols = ['G','K','L','U','W','X']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'FCC')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'BCC'; avec = bra2avec('BCC')
   kSymbols = ['G','H','P','N']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'BCC')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'TET'; avec = bra2avec('TET1')
   kSymbols = ['G','A','M','R','X','Z']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'TET1')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'TET'; avec = bra2avec('TET2')
   kSymbols = ['G','A','M','R','X','Z']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'TET2')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'BCT'; avec = bra2avec('BCT1')
   kSymbols = ['G','M','N','P','X','Z','Z1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'BCT1')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'BCT'; avec = bra2avec('BCT2')
   kSymbols = ['G','N','P','X','Z','S','S1','Y','Y1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'BCT2')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'ORC'; avec = bra2avec('ORC')
   kSymbols = ['G','R','S','T','U','X','Y','Z']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'ORC')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'ORCF'; avec = bra2avec('ORCF1')
   kSymbols = ['G','L','T','Y','Z','A','A1','X','X1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'ORCF1')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'ORCF'; avec = bra2avec('ORCF2')
   kSymbols = ['G','L','X','Y','Z','C','C1','D','D1','H','H1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'ORCF2')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'ORCF'; avec = bra2avec('ORCF3')
   kSymbols = ['G','L','T','Y','Z','A','A1','X','X1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'ORCF3')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'ORCI'; avec = bra2avec('ORCI')
   kSymbols = ['G','R','S','T','W','Z','L','L1','L2','X','X1','Y','Y1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'ORCI')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'ORCC'; avec = bra2avec('ORCC')
   kSymbols = ['G','R','S','T','Y','Z','A','A1','X','X1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'ORCC')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'HEX'; avec = bra2avec('HEX1')
   kSymbols = ['G','A','H','K','L','M']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'HEX1')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'HEX'; avec = bra2avec('HEX2')
   kSymbols = ['G','A','H','K','L','M']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'HEX2')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'RHL'; avec = bra2avec('RHL1')
   kSymbols = ['G','F','L','L1','Z','B','B1','P','P1','P2','Q','X']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'RHL1')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'RHL'; avec = bra2avec('RHL2')
   kSymbols = ['G','F','L','Z','P','P1','Q','Q1']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'RHL2')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'MCL'; avec = bra2avec('MCL')
   kSymbols = ['G','A','C','D','D1','E','X','Y','Y1','Z','H','H1','H2','M','M1','M2']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'MCL')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'MCLC'; avec = bra2avec('MCLC1')
   kSymbols = ['G','N','N1','L','M','Y','Y1','Z','F','F1','F2','F3','I','I1','X','X1','X2']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'MCLC1')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'MCLC'; avec = bra2avec('MCLC3')
   kSymbols = ['G','I','M','N','N1','X','Z','F','F1','F2','H','H1','H2','Y','Y1','Y2','Y3']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'MCLC3')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))

   crystalStructure = 'MCLC'; avec = bra2avec('MCLC5')
   kSymbols = ['G','L','M','N','N1','X','Z','F','F1','F2','H','H1','H2','I','I1','Y','Y1','Y2','Y3']
   print_kvalues(crystalStructure, avec, kSymbols)
   # print('%s' % 'MCLC5')
   # print('%f %f %f' % (avec[0,0],avec[0,1],avec[0,2]))
   # print('%f %f %f' % (avec[1,0],avec[1,1],avec[1,2]))
   # print('%f %f %f' % (avec[2,0],avec[2,1],avec[2,2]))




# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.eig     import Eig
from rescupy.mpidist import Mpidist
from rescupy.mix     import Mix
from rescupy.basis   import Basis
from rescupy.restart import Restart
from rescupy.utils   import asdict as utils_asdict
from rescupy.utils   import init_from_dict

class Solver:
   """
   solver class.
   
   Attributes
   ----------
   basis : basis-object
      Deals with basis related parameters.
   eig : eig-object
      Deals with eigensolver related parameters.
   mix : mix-object
      Deals with mixer related parameters.
   mpi : mpi-object
      Deals with mpi related parameters.
   
   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "basis"   : False,
         "eig"     : False,
         "mix"     : False,
         "mpidist" : False,
         "restart" : False,
      }    
      # default values
      defDict = {
         "basis"   : {},
         "eig"     : {},
         "mix"     : {},
         "mpidist" : {},
         "restart" : {},
      }    
      
      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)

   # optimize solver parameters based on atomSys object
   def optim(self, sobj):
      nval = sobj.atoms.valCharge
      # fix number of target bands
      if self.eig.trgtband is None:
         self.eig.trgtband = [1, int((nval + 1) // 2) + 8]
      # fix number of included bands
      if self.eig.inclband is None:
         self.eig.inclband = [1, int((nval + 1) // 2) + 8]



# -*- coding: utf-8 -*-
"""
Created on 2020-05-12

@author: Vincent Michaud-Rioux
"""

import h5py
import numpy as np
from numpy.polynomial.legendre import leggauss
from scipy.special import spherical_jn as jn
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import load_dcal, load_dcal_var, load_dcal_parameter

class RadFunc:
   """
   radFunc class.
   
   Attributes
   ----------
   funcid : 1D array
      Functional id (LibXC).

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, filename, varname, index=0, ecut=36.75):

      if varname in ['Vna', 'Vlocal', 'Vnl', 'VnlSO']:
         vfld = "vvData"
      if varname in ['Rna', 'Rlocal', 'Rpc']:
         vfld = "rhoData"
      if varname in ['OrbitalSet']:
         vfld = "frData"

      if varname in ['Vna', 'Vlocal','Rna', 'Rlocal', 'Rpc']:
         i = None
      else: 
         i = index

      data, fmt = load_dcal(filename, varname)

      if data is None or len(data) == 0:
         self.rgrid = np.linspace(0, 5, 11)
         self.rwght = np.zeros((11))
         self.rvals = np.zeros((11))
         self.qgrid = np.linspace(0, 5, 11)
         self.qwght = np.zeros((11))
         self.qvals = np.zeros((11))
         self.ecut = max(self.qgrid)
         return
    
      self.rgrid = load_dcal_var(data, 'rrData', fmt, i)
      self.rwght = load_dcal_var(data, 'drData', fmt, i)
      self.rvals = load_dcal_var(data, vfld, fmt, i)

      try:
         self.qgrid = load_dcal_var(data, 'qqData', fmt, i)
         self.qwght = load_dcal_var(data, 'qwData', fmt, i)
         self.qvals = load_dcal_var(data, 'fqData', fmt, i)
      except:
         # self.qgrid, self.qwght = leggauss(int(ecut*10))
         # self.qgrid = (0 * (1 - self.qgrid) + ecut * (1 + self.qgrid)) / 2      
         nq = int(ecut*30)
         l = 0 if type(self) is RadFunc else self.l
         self.qgrid = np.linspace(0, ecut, num = nq + 1)
         self.qwght = np.ones((nq + 1)) / nq * ecut
         self.qwght[0] *= 0.5; self.qwght[-1] *= 0.5
         self.qvals = radialFT(self.rgrid, self.rwght, self.rvals, self.qgrid, l)

      self.ecut = max(self.qgrid)

   def asdict(self):
      return utils_asdict(self)

class Orb(RadFunc):
   def __init__(self, filename, varname, index=0, ecut=36.75):
      i = index
      data, fmt = load_dcal(filename, varname)
      try:
         tmp = load_dcal_parameter(data, 'L', fmt, i)
         self.l = int(tmp)
      except ValueError:
         raise NameError("Couldn't find parameter 'l' in file "+filename+".")
      self.n = -1
      super().__init__(filename, varname, index=index, ecut=ecut)

   def asdict(self):
      return utils_asdict(self)

class KbOrb(Orb):
   def __init__(self, filename, varname, index=0, ecut=36.75):
      i = index
      data, fmt = load_dcal(filename, varname)
      try:    
         tmp = load_dcal_parameter(data, 'halfLxEkbso', fmt, i)
         if tmp is None:
            tmp = load_dcal_parameter(data, 'KBenergy', fmt, i)
         self.kbnrg = float(tmp)
      except ValueError:
         raise NameError("Couldn't find parameter 'KBenergy' in file "+filename+".")
      self.energy = None
      self.kbcos = None

      super().__init__(filename, varname, index=index, ecut=ecut)

   def asdict(self):
      return utils_asdict(self)

class AoOrb(Orb):
   def __init__(self, filename, varname, index=0, ecut=36.75):
      i = index
      data, fmt = load_dcal(filename, varname)
      try:    
         tmp = load_dcal_parameter(data, 'Population', fmt, i)
         self.population = float(tmp)
      except ValueError:
         raise NameError("Couldn't find parameter 'Population' in file "+filename+".")
      self.zeta = -1
      self.energy = None
      self.coulombU = None
      self.exchangeJ = None

      super().__init__(filename, varname, index=index, ecut=ecut)

   def asdict(self):
      return utils_asdict(self)

def radialFT(r, fr, dr, q, l=0):
   """
   Description
   ----------
   Computes the radial Fourier transform.
   
   Parameters
   ----------
   r : 1D array
      Radial grid.
   fr : 1D array
      Radial values.
   dr : 1D array
      Radial integration weights.
   q : 1D array
      Fourier grid.
   l : 1D array
      Principal angular momentum.

   Returns
   ----------
   fq : 1D array
      Radial Fourier transform.
   """
   if l < 0:
      l = 0
   fq = r.reshape((-1,1)) * q.reshape((1,-1))
   fq = jn(l,fq)
   fq = r.reshape((-1,1)) ** 2 * fr.reshape((-1,1)) * fq
   fq = np.sqrt(2 / np.pi) * np.matmul(dr.reshape((1,-1)), fq).flatten()
   return fq
   # reference MATLAB code
   # fq = bsxfun(@times,r(:),q(:)');
   # fq = bsxfun(@times,r(:).^2.*fr(:),calcSphericalBessel(l,fq));
   # fq = sqrt(2/pi)*dr*fq;
 



# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

from scipy.interpolate import interp1d
from scipy.integrate import simps
import numpy as np
from rescupy.orb import RadFunc
from rescupy.pspot import Pspot
from rescupy.aobasis import NaoBasis
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Species:
   """
   species class.
   
   Attributes
   ----------
   label : string
      Species label (e.g. "Si1", "Au_surf").
   charge : float
      Electron charge.
   mass : float
      Atomic mass.
   magmom : float
      Magnetic moment (Cartesian).
   magrad : float
      Magnetic "radius" (for mag. moment integration).
   alphaZ : float
      Long-range energy correction (depends on unit cell volume).
   psp : float
      Pseudopotential.
   aob : float
      Atomic orbital basis.
   rna : float
      Neutral atom valence density (short-range).
   vna : float
      Neutral atom potential (short-range).

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict):
      # required status
      reqDict = {
         "alphaZ": False,
         "charge": False,
         "label" : True ,
         "mass"  : False,
         "magmom": False,
         "magrad": False,
         "path"  : True,
      }
      # default values
      defDict = {
         "alphaZ" : None,
         "charge" : None,
         "label"  : "XX",
         "mass"   : None,
         "magmom" : 0.,
         "magrad" : 1.2,
         "path"   : None,
      }
      
      init_from_dict(self, reqDict, defDict, inpDict)

      # do not update fields if they are found
      if not "psp" in inpDict.keys():
         inpDict["psp"] = {"path":get_path(inpDict)}
      elif not "path" in inpDict["psp"].keys():
         inpDict["psp"]["path"] = get_path(inpDict)
      self.psp = Pspot(inpDict["psp"])
      if not "aob" in inpDict.keys():
         inpDict["aob"] = {"path":get_path(inpDict)}
      elif not "path" in inpDict["aob"].keys():
         inpDict["aob"]["path"] = get_path(inpDict)
      self.aob = NaoBasis(inpDict["aob"], "OrbitalSet")
      self.rna = RadFunc(get_path(inpDict), "Rna")
      self.vna = RadFunc(get_path(inpDict), "Vna")
      self.alphaZ = calcAlphaZ(self.psp.vlc.rgrid, self.psp.vlc.rvals, self.psp.Z)
      if self.charge is None:
         self.charge = self.psp.Z

   def asdict(self):
      return utils_asdict(self)

def get_path(adict):
   if not "path" in adict.keys():
      raise NameError("Input file is missing a value for parameter species.path.")
   return adict["path"]

def calcAlphaZ(rr, vv, Z):
   # rr = spec.psp.Vlocal.rgrid
   # vv = spec.psp.Vlocal.rvals
   # Z = spec.psp.Z
   if rr[0] > 0.:
      tmp = interp1d(rr,vv, kind="cubic", fill_value="extrapolate")
      x = np.array([0.])
      rr = np.concatenate((x, rr))
      vv = np.concatenate((tmp(x), vv))
   integrand = rr ** 2 * vv + Z * rr
   return 4 * np.pi * simps(integrand,rr)
   


# -*- coding: utf-8 -*-
"""
Created on 2020-06-16

@author: Vincent Michaud-Rioux
"""

import h5py
import numpy as np
from math import pi
import scipy.io as sio
import re
import rescupy

def init_from_dict(self, reqDict, defDict, inpDict):
   """
   Initialize object from a dictionary.
   
   If certain attributes are themselves objects, init_from_dict is called recursively.

   Inputs
   ----------
   reqDict : dict
      Dict of keys and whether they are mandatory (True). If so, inpDict must contain 
      a non-trivial value.
   defDict : dict
      Dict of keys and their default values.
   inpDict : dict
      Dict of keys and user defined values.

   Return
   ----------
   self : generic object
      Object storing the attributes and values given by the inputs.

   """
   # Capitalize first letter of classes.
   # https://stackoverflow.com/questions/12410242/python-capitalize-first-letter-only
   fupper = lambda y: re.sub('([a-zA-Z])', lambda x: x.groups()[0].upper(), y, 1)
   for key in defDict.keys():
      if key in inpDict or reqDict[key] is True:
         if not key in inpDict:
            raise NameError('Input dict is missing a value for parameter '+key+'.')
      if isinstance(defDict[key], dict):
         modpath = "rescupy."+key+"."+fupper(key)
         if key in inpDict:
            tmp = eval(modpath+"(inpDict['"+key+"'])")
         else:
            tmp = eval(modpath+"(defDict['"+key+"'])")
      else:
         if key in inpDict:
            tmp = inpDict[key]
         else:
            tmp = defDict[key]
      setattr(self, key, tmp)

def asdict(self):
   """
   Print generic object as a dictionary.
   
   If certain attributes are themselves objects, asdict is called recursively.
   """
   # init dict
   adict = {}
   # loop over attribute
   for key in self.__dict__.keys():
      # get attribute
      attr = getattr(self, key)
      # if it is a list, the first element should be representative, and hence
      # we see whether it is a rescu object (with asdict) or not
      if isinstance(attr, list):
         object_methods = [method_name for method_name in dir(attr[0])
                  if callable(getattr(attr[0], method_name))]
      # special treatment is we have an object list
      if isinstance(attr, list) and "asdict" in object_methods:
         adict[key] = [asdict_core(obj) for obj in attr]
         # adict[key] = adict[key]
      else:
         adict[key] = asdict_core(attr)
   return adict

def asdict_core(attr):
   """
   Print an attribute as a dictionary.
   """
   object_methods = [method_name for method_name in dir(attr)
               if callable(getattr(attr, method_name))]
   if "asdict" in object_methods:
      adict = attr.asdict()
   else:
      adict = attr
   if isinstance(adict, np.ndarray):
      adict = adict.flatten().tolist()
   return adict

def load_dcal(filename, varname=None):
   try:
      data = sio.loadmat(filename)
      data = data['data'][0]
      if not varname is None:
         data = data[varname][0]
      fmt = 'mat'
   except:
      data = h5py.File(filename, 'r')
      data = data['data']
      if not varname is None:
         data = data[varname]
      fmt = 'h5'
   if len(data) == 0:
      data = None
   return data, fmt

def load_dcal_var(data, varname, fmt, index):
   i = index
   if fmt == 'mat':
      if index is None:
         i = 0
      var = data[0][varname][i].squeeze()
   elif index is None:
      var = data[varname][0:].squeeze()
   else:
      var = data[data[varname][i][0]][0:].flatten()
   return var

def load_dcal_parameter(data, varname, fmt, index):
   i = index
   if fmt == 'mat':
      data = data[0]
      flds = list(data['Parameter'][i].dtype.names)
      if varname in flds:
         parameter = data['Parameter'][i][varname][0][0]
      else:
         parameter = None
   else:
      tmp = data[data['Parameter'][i][0]]
      flds = tmp.keys()
      if varname in flds:
         parameter = tmp[varname][0]
      else:
         parameter = None
   return parameter

def set_units(self, units):
   """
   Sets units to either atomic or si.
   """
   units = units.lower()
   if not units in ["atomic", "si"]:
      raise NameError("Unit system "+units+" not recognized.")
   for key in self.__dict__.keys():
      attr = getattr(self, key)
      object_methods = [method_name for method_name in dir(attr)
                  if callable(getattr(attr, method_name))]
      if "set_units" in object_methods:
         attr.set_units(units)
         setattr(self, key, attr)
         continue
      if len(key) < 5: 
         continue
      if key[-5:] != "Units":
         continue
      attr = getattr(self, key[:-5])
      unit = getattr(self, key)
      attr, unit = convert_units(attr, unit, units)
      setattr(self, key[:-5], attr)
      setattr(self, key, unit)

def convert_units(q, qUnits, units):
   """
   Convert quantity "q" with units "units" to system "units".
   """
   udict = get_units()
   qUnits = qUnits.lower()
   if qUnits in ["ang", "angstrom", "bohr"]:
      uq = udict[qUnits]
      if units == "atomic": 
         newunit = "bohr"
      if units == "si": 
         newunit = "ang"
   elif qUnits in ["invang", "invbohr"]:
      uq = udict[qUnits]
      if units == "atomic": 
         newunit = "invbohr"
      if units == "si": 
         newunit = "invang"
   elif qUnits in ["ev", "ha", "hartree"]:
      uq = udict[qUnits]
      if units == "atomic": 
         newunit = "ha"
      if units == "si": 
         newunit = "ev"
   elif qUnits in ["ev/ang", "ha/bohr"]:
      uq = udict[qUnits]
      if units == "atomic": 
         newunit = "ha/bohr"
      if units == "si": 
         newunit = "ev/ang"
   elif qUnits in ["ev/ang3", "ha/bohr3"]:
      uq = udict[qUnits]
      if units == "atomic": 
         newunit = "ha/bohr3"
      if units == "si": 
         newunit = "ev/ang3"
   elif qUnits in ["invev", "invha"]:
      uq = udict[qUnits]
      if units == "atomic": 
         newunit = "invha"
      if units == "si": 
         newunit = "invev"
   else:
      raise NameError("Units dimension of "+qUnits+" not recognized.")
   if q is None:
      return q, newunit
   utarg = udict[newunit]
   if isinstance(q, list):
      q = [q0 * uq / utarg for q0 in q]
   else:
      q = q * uq / utarg
   return q, newunit

def read_field(filename, fieldname):
   """
   Read a field from an HDF5 file.

   Parameters: 
      filename (str): Path the the HDF5 file. For example, "rescu_scf_out.h5".

      fieldname (str): Path of the field in the HDF5 file. For example, "potential/effective".

   Returns: 
      fld (ndarray): 3D numpy array containing the field.
   """
   units = get_units()
   f = h5py.File(filename, mode='r')
   fld = f[fieldname][0:]
   fld = np.transpose(fld, [i for i in range(fld.ndim-1,-1,-1)])
   if re.match('potential', fieldname):
      fld = fld * units['ha']
   elif re.match('density', fieldname):
      fld = fld / units['bohr'] ** 3
   else:
      raise NameError('Unknown field type.')
   return fld

def get_units():
   """
   Function that creates a dictionary containing units.
   """
   # CODATA 2018 taken from
   # https://physics.nist.gov/cuu/Constants/index.html
   u = {"_c": 299792458.,            # Exact
         "_mu0": 4.0e-7 * pi,        # Exact
         "_grav": 6.67430e-11,       # +/- 0.000_15e-11
         "_hplanck": 6.62607015e-34, # Exact
         "_e": 1.602176634e-19,      # Exact
         "_me": 9.1093837015e-31,    # +/- 0.000_000_0028e-31
         "_mp": 1.67262192369e-27,   # +/- 0.000_000_000_51e-27
         "_nav": 6.02214076e23,      # Exact
         "_k": 1.380649e-23,         # Exact
         "_amu": 1.66053906660e-27}  # +/- 0.000_000_000_50e-27

   # derived from the CODATA values
   u["_eps0"] = (1 / u["_mu0"] / u["_c"]**2)  # permittivity of vacuum
   u["_hbar"] = u["_hplanck"] / (2 * pi)  # Planck constant / 2pi, J s

   u["ang"] = u["angstrom"] = 1.0
   u["nm"] = 10.0
   u["bohr"] = (4e10 * pi * u["_eps0"] * u["_hbar"]**2 /
               u["_me"] / u["_e"]**2)  # Bohr radius
   u["invang"] = 1. / u["ang"]
   u["invbohr"] = 1. / u["bohr"]

   u["ev"] = 1.0
   u["hartree"] = (u["_me"] * u["_e"]**3 / 16 / pi**2 /
                  u["_eps0"]**2 / u["_hbar"]**2)
   u["rydberg"] = 0.5 * u["hartree"]
   u["ry"] = u["rydberg"]
   u["ha"] = u["hartree"]
   # u["kj"] = 1000.0 / u["_e"]
   # u["kcal"] = 4.184 * u["kj"]
   # u["mol"] = u["_nav"]

   u["invev"] = 1.0 / u["ev"]
   u["invha"] = 1.0 / u["ha"]

   u["ev/ang"] = 1.0
   u["ha/bohr"] = 1.0 * u["ha"] / u["bohr"]

   u["ev/ang3"] = 1.0
   u["ha/bohr3"] = 1.0 * u["ha"] / (u["bohr"] ** 3)

   # u["second"] = 1e10 * sqrt(u["_e"] / u["_amu"])
   # u["fs"] = 1e-15 * u["second"]

   # u["kb"] = u["_k"] / u["_e"]  # Boltzmann constant, eV/K

   # u["pascal"] = (1 / u["_e"]) / 1e30  # J/m^3
   # u["gpa"] = 1e9 * u["pascal"]

   # u["debye"] = 1.0 / 1e11 / u["_e"] / u["_c"]
   # u["alpha"] = (u["_e"]**2 / (4 * pi * u["_eps0"]) /
   #             u["_hbar"] / u["_c"])  # fine structure constant
   # u["invcm"] = (100 * u["_c"] * u["_hplanck"] /
   #             u["_e"])  # cm^-1 energy unit

   # # Derived atomic units that have no assigned name:
   # # atomic unit of time, s:
   # u["_aut"] = u["_hbar"] / (u["alpha"]**2 * u["_me"] * u["_c"]**2)
   # # atomic unit of velocity, m/s:
   # u["_auv"] = u["_e"]**2 / u["_hbar"] / (4 * pi * u["_eps0"])
   # # atomic unit of force, N:
   # u["_auf"] = u["alpha"]**3 * u["_me"]**2 * u["_c"]**3 / u["_hbar"]
   # # atomic unit of pressure, Pa:
   # u["_aup"] = u["alpha"]**5 * u["_me"]**4 * u["_c"]**5 / u["_hbar"]**3

   # u["aut"] = u["second"] * u["_aut"]

   # # SI units
   # u["m"] = 1e10 * u["ang"]  # metre
   # u["kg"] = 1. / u["_amu"]  # kilogram
   # u["s"] = u["second"]  # second
   # u["a"] = 1.0 / u["_e"] / u["s"]  # ampere
   # # derived
   # u["j"] = u["kj"] / 1000  # Joule = kg * m**2 / s**2
   # u["c"] = 1.0 / u["_e"]  # Coulomb = A * s

   return u

def get_chemical_symbols():
   """Get a list of atomic species.

   Source
   -------
   Atomic Simulation Environment
   https://wiki.fysik.dtu.dk/ase/index.html
   commit 68f2a4c1e4c492654c32eb7e2bcd3a43a0dedf00 (origin/master, origin/HEAD, master)
   Merge: 22bbcf87e e13cd3184
   Author: Ask Hjorth Larsen <asklarsen@gmail.com>
   Date:   Tue May 5 17:08:04 2020 +0000
   """
   chemical_symbols = [
      # 0
      'X',
      # 1
      'H', 'He',
      # 2
      'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
      # 3
      'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar',
      # 4
      'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
      'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',
      # 5
      'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
      'In', 'Sn', 'Sb', 'Te', 'I', 'Xe',
      # 6
      'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
      'Ho', 'Er', 'Tm', 'Yb', 'Lu',
      'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi',
      'Po', 'At', 'Rn',
      # 7
      'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk',
      'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr',
      'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Nh', 'Fl', 'Mc',
      'Lv', 'Ts', 'Og']
   return chemical_symbols


# -*- coding: utf-8 -*-
"""
Created on 2020-06-16

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Restart:
   """
   restart class.
   
   Attributes
   ----------
   densityPath : string

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "densityPath": False,
         "DMkPath"    : False,
         "DMRPath"    : False,
      }    
      # default values
      defDict = {
         "densityPath": "",
         "DMkPath"    : "",
         "DMRPath"    : "",
      }    
      
      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)


# -*- coding: utf-8 -*-
"""
Created on 2020-05-12

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import set_units as utils_set_units

class Pop:
   """
   pop class.
   
   Attributes
   ----------
   type : string
      Occupancy scheme (e.g. Fermi-Dirac, Gaussian)
   blochl : logical
      Blochl correction in the tetrahedron scheme.
   mpn : integer
      Methfessel-Paxton scheme order.
   sigma : float
      Smearing parameter.

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "type"      : False,
         "blochl"    : False,
         "mpn"       : False,
         "sigma"     : False,
         "sigmaUnits": False,
      }
      # default values
      defDict = {
         "type"      : "ga",
         "blochl"    : False,
         "mpn"       : 1,
         "sigma"     : 0.1,
         "sigmaUnits": "ev",
      }
      
      init_from_dict(self, reqDict, defDict, inpDict)

      self.type = long2short(self.type)

      if not self.type in ['fx','fd','ga','mp','tm']: 
         raise NameError('Invalid pop.type value: '+self.type+'.')
      
      if self.type is 'fx':
         self.sigma = 0.0
         
      if not self.type is 'mp':
         self.mpn = 0
         
   def asdict(self):
      return utils_asdict(self)

   def set_units(self, units):
      return utils_set_units(self, units)

def long2short(long):
   if long == 'fixed':
      return 'fx'
   if long == 'fermi-dirac':
      return 'fd'
   if long == 'gauss':
      return 'ga'
   if long == 'methfessel-paxton':
      return 'mp'
   if long == 'tetrahedron':
      return 'tm'
   return long



# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
import matplotlib.pyplot as plt
from rescupy.atomsys import AtomSys
from rescupy.cmd     import Cmd
from rescupy.solver  import Solver
from rescupy.kmesh   import increase_ksampling
from rescupy.utils   import asdict as utils_asdict
from rescupy.utils   import init_from_dict
from rescupy.utils   import read_field
from rescupy.utils   import set_units as utils_set_units
import json
import os 

class Rescu:
   """
   rescu class.
   
   Attributes
   ----------
   solver : solver object           
      Object containing solver related parameters.
   atomSys : atomSys object           
      Object containing atomSys related parameters.
   
   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict={}, filename=None):
      if not filename is None:
         self.read_output(filename, units="si")
         return

      # atomic system
      if not 'atomSys' in inpDict:
         NameError('"atomSys" parameters not found in input file.')
      self.atomSys = AtomSys(inpDict['atomSys'])

      # solver
      if 'solver' in inpDict:
         tmp = inpDict['solver']
      else:
         tmp = {}
      tmp = Solver(tmp)
      tmp.optim(self.atomSys)
      self.solver = tmp

      # rescu commands
      if 'cmd' in inpDict:
         tmp = inpDict['cmd']
      else:
         tmp = {}
      self.cmd = Cmd(tmp)

   def asdict(self):
      return utils_asdict(self)

   def conv_ksampling(self, etol=1e-2):
      if not "scfcmd" in self.__dict__.keys():
         NameError("Error in conv_ksampling: scfcmd not set. It can be set calling the set_cmd method.")
      basedict = self.asdict()
      detot = max(1., etol*10.)
      count = 0
      kref = self.atomSys.kmesh.grid; k1 = kref
      etot = []; kall = []
      print("%20s | %20s | %20s" % ("ksampling", "total energy (ev)", "delta energy (ev)"))
      while (abs(detot) > etol):
         hdf5file = "conv_ksampling_"+str(count)+".h5"
         jsonfile = "conv_ksampling_"+str(count)+".json"
         basedict["atomSys"]["kmesh"] = {"grid":k1, "gammaCentered":True}
         basedict["solver"]["restart"]["DMRPath"] = "conv_ksampling_"+str(count-1)+".h5"
         rscobj = Rescu(basedict)
         rscobj.write_input(jsonfile)
         rscobj.cmd.scf(jsonfile)
         os.system("mv rescu_scf_out.h5 "+hdf5file)
         os.system("mv rescu_scf_out.json "+jsonfile)
         rscobj.read_output(jsonfile)
         etot.append(rscobj.atomSys.energy.etot)
         kall.append(rscobj.atomSys.kmesh.grid)
         if count > 0:
            detot = etot[-1] - etot[-2]
         else:
            detot = etot[-1]
         print("%6d %6d %6d | %+20.8f | %+20.8f" % (k1[0], k1[1], k1[2], etot[-1], detot))
         k1 = increase_ksampling(k1, kref)
         count += 1
      detot = [e - etot[-1] for e in etot]
      print("%20s | %20s | %20s" % ("ksampling", "total energy (ev)", "energy error (ev)"))
      for i in range(0, len(etot)):
         print("%6d %6d %6d | %+20.8f | %+20.8f" % (kall[i][0], kall[i][1], kall[i][2], etot[i], detot[i]))

   def conv_resolution(self, etol=1e-2):
      if not "scfcmd" in self.__dict__.keys():
         NameError("Error in conv_resolution: scfcmd not set. It can be set calling the set_cmd method.")
      basedict = self.asdict()
      detot = max(1., etol*10.)
      count = 0
      res1 = self.atomSys.cell.lres
      etot = []; res = []
      print("%20s | %20s | %20s" % ("resolution (ang)", "total energy (ev)", "delta energy (ev)"))
      while (abs(detot) > etol):
         hdf5file = "conv_resolution_"+str(count)+".h5"
         jsonfile = "conv_resolution_"+str(count)+".json"
         basedict["atomSys"]["cell"]["grid"] = None
         basedict["atomSys"]["cell"]["lres"] = res1
         basedict["solver"]["restart"]["DMRPath"] = "conv_resolution_"+str(count-1)+".h5"
         rscobj = Rescu(basedict)
         rscobj.write_input(jsonfile)
         rscobj.cmd.scf(jsonfile)
         os.system("mv rescu_scf_out.h5 "+hdf5file)
         os.system("mv rescu_scf_out.json "+jsonfile)
         rscobj.read_output(jsonfile)
         etot.append(rscobj.atomSys.energy.etot)
         res.append(rscobj.atomSys.cell.lres)
         if count > 0:
            detot = etot[-1] - etot[-2]
         else:
            detot = etot[-1]
         print("%20.8f | %+20.8f | %+20.8f" % (res[-1], etot[-1], detot))
         res1 *= 0.9
         count += 1
      detot = [e - etot[-1] for e in etot]
      print("%20s | %20s | %20s" % ("resolution (ang)", "total energy (ev)", "energy error (ev)"))
      for i in range(0, len(etot)):
         print("%20.8f | %+20.8f | %+20.8f" % (res[i], etot[i], detot[i]))

   def plot_field(self, field, avg):
      """
      Plot the average of a field. 

      Parameters: 

         field (str): Path in the HDF5 filed pointed to by self.solver.restart.densityPath.

         avg (tuple): Length-2 tuple containing the axes along which to average the field.

      Returns:
         fig (figure): Pyplot figure showing the averaged field.  
      """
      if len(avg) != 2:
         raise NameError("Input avg must be of length 2.")
      if avg[0] == avg[1]:
         raise NameError("Input avg must have distinct entries.")
      axis = -1
      for d in range(3):
         if not d in avg:
            axis = d; break
      if not axis in [0,1,2]:
         raise NameError("Invalid input avg.")
      filename = self.solver.restart.densityPath
      field = read_field(filename, field)
      vavg = np.mean(field, axis=avg)
      x = self.atomSys.cell.get_grid(axis=axis)[:,axis]
      lx = self.atomSys.cell.get_length(axis=axis)
      fig = plt.figure()
      plt.plot(x, vavg)
      plt.xticks([0, lx])
      plt.grid(axis='x')
      plt.xlabel("position ("+self.atomSys.cell.avecUnits+")")
      plt.ylabel("field average (eV)")
      plt.show()
      return fig

   def set_units(self, units):
      return utils_set_units(self, units)

   def set_cmd(self, inpDict = {}):
      self.cmd = Cmd(inpDict)

   def write_input(self, filename, units="atomic"):
      self.set_units(units)
      adict = self.asdict()
      fid = open(filename, "w")
      json.dump(adict, fid, indent=2, sort_keys=True)
      fid.close()

   def read_output(self, filename, units="si"):
      fid = open(filename, "r")
      adict = json.load(fid)
      fid.close()
      self.__init__(adict)
      self.set_units(units)
